/*  FRAG.C
 *  demonstrates failure of Microsoft's _fmalloc()
 *  under certain conditions, and that mem_alloc()
 *  does not fail under the same conditions
 *  last mod.: 11-FEB-91
 */

#include <STDIO.H>
#include <STDLIB.H>
#include <MALLOC.H>
#include <CONIO.H>
#include <FCNTL.H>
#include <IO.H>

#include <L_MEM.H>

#define DOLPHIN FALSE    /*  switches between _fmalloc() and mem_alloc()  */

#if DOLPHIN
/*  memory log not used in this program so:  */
#include <NOMEMLOG.C>
/*  and compile with /NOE linker switch  */
char *program = "FRAG_D";       /*  rename FRAG.EXE to FRAG_D.EXE  */
#else
char *program = "FRAG_M";       /*  rename FRAG.EXE to FRAG_M.EXE  */
#endif

#define N       40       /*  number of pointers  */

void far * ptr[N];

void main (int argc,char **argv)
{
Uint i, j, k, size, n;
char a[4];
#if DOLPHIN
int err_flag;
#endif

printf("\nUse \"%s [random_seed [mem_alloc_strategy]]\"."
    "\nmem_alloc_strategy = 0, 1 or 2"
    "\nThis program allocates blocks of even random size"
    "\nin the range 19,000 through 65,534 bytes."
    "\nWhen an attempt to allocate a block of a given size fails, randomly"
    "\nchosen blocks are freed until a block of that size can be allocated."
    "\n\"+\" is displayed when an allocation occurs."
    "\n\"!\" is displayed when an attempt to allocate fails"
    "\n      followed by the number of currently allocated blocks."
    "\n\"-\" is displayed when a block is freed.",program);
printf("\nThe program continues until an allocation fails even though"
    "\nall previously allocated blocks have been freed.\n");

#if DOLPHIN
printf("\nPress a key to begin (and control-break to quit) ...\n");
#else
printf("\nPress a key to begin ...\n");
#endif
getch();

srand(atoi(argv[1]));
n = ( argc < 3 ? 0 : argv[2][0]-'0' );
set_mem_alloc_strategy( n < 3 ? n : 0 );

#if DOLPHIN
/*  trap any critical error that occurs  */
crit_err_termination_on();
#endif

/*  initialize pointers to NULL  */
for ( i=0; i<N; i++ )
    ptr[i] = (void far *)NULL;

while ( TRUE )
    {
    i = rand()%N;       /*  pick an unused pointer at random  */
    if ( ptr[i] == (void far *)NULL )
        {
        /*  randomly pick an even size in the
         *  range 19000 through 65,534 bytes
         */
        size = 0;
        while ( size < 19000 )
            size = 2*(Uint)rand();
        /*  attempt to allocate a block of this size  */
        while ( TRUE )
            {
            #if DOLPHIN     /*  use Dolphin Software's mem_alloc()  */
            if ( ( ptr[i] = mem_alloc((Ulong)size,&err_flag) )
                != (void far *)NULL )
            #else           /*  use Microsoft's _fmalloc()  */
            if ( ( ptr[i] = _fmalloc(size) ) != (void far *)NULL )
            #endif
                {
                printf("+");    /*  block allocated  */
                break;          /*  go get another unused pointer  */
                }
            else
                {
                printf("!");    /*  allocation attempt failed  */
                /*  see if any blocks available to free  */
                n = 0;
                for ( k=0; k<N; k++ )
                    n += ( ptr[k] != (void far *)NULL );
                /*  display number of blocks currently allocated  */
                printf(itoa(n,a,10));
                if ( n )       /*  if there are blocks left to free  */
                    {
                    /*  free a randomly selected block  */
                    while ( TRUE )
                        {
                        j = rand()%N;
                        if ( ptr[j] != (void far *)NULL )
                            {
                            #if DOLPHIN
                            mem_free(ptr[j],&err_flag);
                            #else
                            _ffree(ptr[j]);
                            #endif
                            printf("-");
                            /*  show pointer as not in use  */
                            ptr[j] = (void far *)NULL;
                            break;  /*  go try the allocation again  */
                            }
                        }
                    }
                else        /*  there are no blocks left to free  */
                    {
                    printf("\nA block of %u bytes could not be allocated, "
                        "even though"
                        "\nall previously allocated blocks had been freed.",
                        size);
                    /*  set memory allocation strategy to normal value  */
                    set_mem_alloc_strategy(FIRST_FIT);
                    exit(1);
                    }
                }
            }
        }
    }
}
